from .cqc_handler import CQCHandler
from .cqc_connection import CQCConnection
from .cqc_mix import CQCMix, CQCVariable, CQCMixConnection, mix_qubit
from .cqc_to_file import CQCToFile
from .qubit import qubit
from .util import (
    ProgressBar,
    CQCGeneralError,
    CQCNoQubitError,
    CQCUnsuppError,
    CQCTimeoutError,
    CQCInuseError,
    CQCUnknownError,
    QubitNotActiveError,
)



#
# Copyright (c) 2017, Stephanie Wehner and Axel Dahlberg
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
# 1. Redistributions of source code must retain the above copyright
#    notice, this list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright
#    notice, this list of conditions and the following disclaimer in the
#    documentation and/or other materials provided with the distribution.
# 3. All advertising materials mentioning features or use of this software
#    must display the following acknowledgement:
#    This product includes software developed by Stephanie Wehner, QuTech.
# 4. Neither the name of the QuTech organization nor the
#    names of its contributors may be used to endorse or promote products
#    derived from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDER ''AS IS'' AND ANY
# EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER BE LIABLE FOR ANY
# DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import logging
import warnings
from cqc.cqcHeader import (
    CQC_CMD_NEW,
    CQC_CMD_I,
    CQC_CMD_X,
    CQC_CMD_Y,
    CQC_CMD_Z,
    CQC_CMD_T,
    CQC_CMD_H,
    CQC_CMD_K,
    CQC_CMD_ROT_X,
    CQC_CMD_ROT_Y,
    CQC_CMD_ROT_Z,
    CQC_CMD_CNOT,
    CQC_CMD_CPHASE,
    CQC_CMD_MEASURE,
    CQC_CMD_MEASURE_INPLACE,
    CQC_CMD_RESET,
    CQC_CMD_RELEASE,
)
from .util import (
    CQCGeneralError,
    CQCUnsuppError,
    QubitNotActiveError,
)


class qubit:
    """
    A qubit.
    """

    def __init__(self, cqc, notify=True, block=True, createNew=True, q_id=None, entInfo=None):
        """
        Initializes the qubit. The cqc connection must be given.
        If notify, the return message is received before the method finishes.
        createNew is set to False when we receive a qubit.

        - **Arguments**

            :cqc:         The CQCconnection used
            :nofify:     Do we wish to be notified when done.
            :block:         Do we want the qubit to be blocked
            :createNew:     If NEW-message should be sent, used internally
            :q_id:         Qubit id, used internally if createNew
            :entInfo:     Entanglement information, if qubit is part of EPR-pair
        """
        # Cqc connection
        self._cqc = cqc

        self.notify = cqc.notify

        # Check if the cqc connection was openened using a 'with' statement
        # If not, raise a deprecation warning
        if not self._cqc._opened_with_with:
            logging.info(
                "You should open CQCConnection in a context, i.e. using 'with CQCConnection(...) as cqc:'. "
                "Then qubits will be automatically released by the end of the program, independently of what happens. "
                "For more information, see https://softwarequtech.github.io/SimulaQron/html/PythonLib.html"
            )

        # Whether the qubit is active. Will be set in the first run
        self._active = None

        if createNew:
            # print info
            logging.debug("App {} tells CQC: 'Create qubit'".format(self._cqc.name))

            # Create new qubit at the cqc server
            # TODO how to handle pending headers
            self._cqc.commit_command(0, CQC_CMD_NEW, notify=notify, block=block)
            
            # Get qubit id
            try:
                self._qID = self._cqc.new_qubitID()
            except AttributeError:
                raise CQCGeneralError("Didn't receive the qubit id")
            
            # Activate qubit
            self._set_active(True)
            if notify and self.notify:
                message = self._cqc.readMessage()
                self._cqc.print_CQC_msg(message)
        
        else:
            self._qID = q_id
            self._set_active(False)  # Why?

        # Entanglement information
        self._set_entanglement_info(entInfo)

    def __str__(self):
        if self.active:
            return "Qubit at the node {}".format(self._cqc.name)
        else:
            return "Not active qubit"

    @property
    def entanglement_info(self):
        return self._entInfo

    def _set_entanglement_info(self, ent_info):
        self._entInfo = ent_info

        # Lookup remote entangled node
        self._remote_entNode = None
        if self._entInfo:
            ip = self._entInfo.node_B
            port = self._entInfo.port_B
            try:
                for node in self._cqc._cqcNet.hostDict.values():
                    if (node.ip == ip) and (node.port == port):
                        self._remote_entNode = node.name
                        break
            except AttributeError:
                self._remote_entNode = None

    @property
    def active(self):
        return self._active

    @property
    def remote_entangled_node(self):
        return self._remote_entNode

    def get_entInfo(self):
        warnings.warn("get_entInfo is deprecated, use the property entanglement_info instead",
                      DeprecationWarning)
        return self._entInfo

    def print_entInfo(self):
        warnings.warn("print_entInfo is deprecated",
                      DeprecationWarning)
        if self.entanglement_info:
            print(self.entanglement_info)
        else:
            print("No entanglement information")

    def set_entInfo(self, entInfo):
        warnings.warn("set_entInfo is deprecated, use _set_entanglement_info instead",
                      DeprecationWarning)
        self._set_entanglement_info(entInfo)

    def is_entangled(self):
        if self._entInfo:
            return True
        return False

    def get_remote_entNode(self):
        warnings.warn("get_remote_entNode is deprecated, use the propery remote_entangled_node instead",
                      DeprecationWarning)
        return self.remote_entangled_node

    def check_active(self):
        """
        Checks if the qubit is active
        """
        if not self.active:
            raise QubitNotActiveError("Qubit is not active")

    def _set_active(self, be_active):

        # Check if not already new state
        if self._active == be_active:
            return

        if be_active:
            self._cqc.active_qubits.append(self)
        else:
            if self in self._cqc.active_qubits:
                self._cqc.active_qubits.remove(self)
        self._active = be_active

    def _single_qubit_gate(self, command, notify, block):
        """
        Performs a single qubit gate specified by the command, called in I(), X(), Y() etc
        :param command: the identifier of the command, as specified in cqcHeader.py
        :param notify: Do we wish to be notified when done
        :param block: Do we want the qubit to be blocked
        """
        # check if qubit is active
        self.check_active()

        notify = notify and self.notify

        self._cqc.put_command(
            qID=self._qID,
            command=command,
            notify=notify,
            block=block,
        )

    def I(self, notify=True, block=True):
        """
        Performs an identity gate on the qubit.
        If notify, the return message is received before the method finishes.

        - **Arguments**

            :nofify:     Do we wish to be notified when done.
            :block:         Do we want the qubit to be blocked
        """
        self._single_qubit_gate(CQC_CMD_I, notify, block)

    def X(self, notify=True, block=True):
        """
        Performs a X on the qubit.
        If notify, the return message is received before the method finishes.

        - **Arguments**

            :nofify:     Do we wish to be notified when done.
            :block:         Do we want the qubit to be blocked
        """
        self._single_qubit_gate(CQC_CMD_X, notify, block)

    def Y(self, notify=True, block=True):
        """
        Performs a Y on the qubit.
        If notify, the return message is received before the method finishes.

        - **Arguments**

            :nofify:     Do we wish to be notified when done.
            :block:         Do we want the qubit to be blocked
        """
        self._single_qubit_gate(CQC_CMD_Y, notify, block)

    def Z(self, notify=True, block=True):
        """
        Performs a Z on the qubit.
        If notify, the return message is received before the method finishes.

        - **Arguments**

            :nofify:     Do we wish to be notified when done.
            :block:         Do we want the qubit to be blocked
        """
        self._single_qubit_gate(CQC_CMD_Z, notify, block)

    def T(self, notify=True, block=True):
        """
        Performs a T gate on the qubit.
        If notify, the return message is received before the method finishes.

        - **Arguments**

            :nofify:     Do we wish to be notified when done.
            :block:         Do we want the qubit to be blocked
        """
        self._single_qubit_gate(CQC_CMD_T, notify, block)

    def H(self, notify=True, block=True):
        """
        Performs a Hadamard on the qubit.
        If notify, the return message is received before the method finishes.

        - **Arguments**

            :nofify:     Do we wish to be notified when done.
            :block:         Do we want the qubit to be blocked
        """
        self._single_qubit_gate(CQC_CMD_H, notify, block)

    def K(self, notify=True, block=True):
        """
        Performs a K gate on the qubit.
        If notify, the return message is received before the method finishes.

        - **Arguments**

            :nofify:     Do we wish to be notified when done.
            :block:         Do we want the qubit to be blocked
        """
        self._single_qubit_gate(CQC_CMD_K, notify, block)

    def _single_gate_rotation(self, command, step, notify, block):
        """
        Perform a rotation on a qubit
        :param command: the rotation qubit command as specified in cqcHeader.py
        :param step: Determines the rotation angle in steps of 2*pi/256
        :param notify: Do we wish to be notified when done
        :param block: Do we want the qubit to be blocked
        :return:
        """
        # check if qubit is active
        self.check_active()

        notify = notify and self.notify

        self._cqc.put_command(
            qID=self._qID,
            command=command,
            step=step,
            notify=notify,
            block=block,
        )

    def rot_X(self, step, notify=True, block=True):
        """
        Applies rotation around the x-axis with the angle of step*2*pi/256 radians.
        If notify, the return message is received before the method finishes.

        - **Arguments**

            :step:         Determines the rotation angle in steps of 2*pi/256
            :nofify:     Do we wish to be notified when done.
            :block:         Do we want the qubit to be blocked
        """
        self._single_gate_rotation(CQC_CMD_ROT_X, step, notify, block)

    def rot_Y(self, step, notify=True, block=True):
        """
        Applies rotation around the y-axis with the angle of step*2*pi/256 radians.
        If notify, the return message is received before the method finishes.

        - **Arguments**

            :step:         Determines the rotation angle in steps of 2*pi/256
            :nofify:     Do we wish to be notified when done.
            :block:         Do we want the qubit to be blocked
        """
        self._single_gate_rotation(CQC_CMD_ROT_Y, step, notify, block)

    def rot_Z(self, step, notify=True, block=True):
        """
        Applies rotation around the z-axis with the angle of step*2*pi/256 radians.
        If notify, the return message is received before the method finishes.

        - **Arguments**

            :step:         Determines the rotation angle in steps of 2*pi/256
            :nofify:     Do we wish to be notified when done.
            :block:         Do we want the qubit to be blocked
        """
        self._single_gate_rotation(CQC_CMD_ROT_Z, step, notify, block)

    def _two_qubit_gate(self, command, target, notify, block):
        """
        Perform a two qubit gate on the qubit
        :param command: the two qubit gate command as specified in cqcHeader.py
        :param target: The target qubit
        :param notify: Do we wish to be notified when done
        :param block: Do we want the qubit to be blocked
        """
        # check if qubit is active
        self.check_active()
        target.check_active()

        if self._cqc != target._cqc:
            raise CQCUnsuppError("Multi qubit operations can only operate on qubits in the same process")

        if self == target:
            raise CQCUnsuppError("Cannot perform multi qubit operation where control and target are the same")

        notify = notify and self.notify

        self._cqc.put_command(
            qID=self._qID,
            command=command,
            notify=notify,
            block=block,
            xtra_qID=target._qID,
        )

    def cnot(self, target, notify=True, block=True):
        """
        Applies a cnot onto target.
        Target should be a qubit-object with the same cqc connection.
        If notify, the return message is received before the method finishes.

        - **Arguments**

            :target:     The target qubit
            :nofify:     Do we wish to be notified when done.
            :block:         Do we want the qubit to be blocked
        """
        self._two_qubit_gate(CQC_CMD_CNOT, target, notify, block)

    def cphase(self, target, notify=True, block=True):
        """
        Applies a cphase onto target.
        Target should be a qubit-object with the same cqc connection.
        If notify, the return message is received before the method finishes.

        - **Arguments**

            :target:     The target qubit
            :nofify:     Do we wish to be notified when done.
            :block:         Do we want the qubit to be blocked
        """
        self._two_qubit_gate(CQC_CMD_CPHASE, target, notify, block)

    def measure(self, inplace=False, block=True):
        """
        Measures the qubit in the standard basis and returns the measurement outcome.
        If now MEASOUT message is received, None is returned.
        If inplace=False, the measurement is destructive and the qubit is removed from memory.
        If inplace=True, the qubit is left in the post-measurement state.

        - **Arguments**

            :inplace:     If false, measure destructively.
            :block:         Do we want the qubit to be blocked
        """
        # check if qubit is active
        self.check_active()

        if inplace:
            command = CQC_CMD_MEASURE_INPLACE
        else:
            command = CQC_CMD_MEASURE
            # Set qubit to non active so the user can receive helpful errors during compile time 
            # if this qubit is used after this measurement
            self._set_active(False)

        self._cqc.put_command(
            qID=self._qID,
            command=command,
            notify=False,
            block=block,
        )

        if self._cqc.pend_messages:
            return None
        else:
            return self._cqc.return_meas_outcome()

    def reset(self, notify=True, block=True):
        """
        Resets the qubit.
        If notify, the return message is received before the method finishes.

        - **Arguments**

            :nofify:     Do we wish to be notified when done.
            :block:         Do we want the qubit to be blocked
        """
        # check if qubit is active
        self.check_active()

        notify = notify and self.notify

        self._cqc.put_command(
            qID=self._qID,
            command=CQC_CMD_RESET,
            notify=notify,
            block=block,
        )

    def release(self, notify=True, block=True):
        """
        Release the current qubit
        :param notify: Do we wish to be notified when done
        :param block: Do we want the qubit to be blocked
        :return:
        """

        notify = notify and self.notify

        self._set_active(False)

        self._cqc.put_command(
            qID=self._qID,
            command=CQC_CMD_RELEASE,
            notify=notify,
            block=block,
        )

    def getTime(self, block=True):
        """
        Returns the time information of the qubit.
        If no INF_TIME message is received, None is returned.

        - **Arguments**

            :block:         Do we want the qubit to be blocked
        """
        # check if qubit is active
        self.check_active()

        # print info
        logging.debug("App {} tells CQC: 'Return time-info of qubit with ID {}'".format(self._cqc.name, self._qID))

        self._cqc.sendGetTime(self._qID, notify=0, block=int(block))

        # Return time-stamp
        message = self._cqc.readMessage()
        try:
            otherHdr = message[1]
            return otherHdr.datetime
        except AttributeError:
            return None



#
# Copyright (c) 2017, Stephanie Wehner and Axel Dahlberg
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
# 1. Redistributions of source code must retain the above copyright
#    notice, this list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright
#    notice, this list of conditions and the following disclaimer in the
#    documentation and/or other materials provided with the distribution.
# 3. All advertising materials mentioning features or use of this software
#    must display the following acknowledgement:
#    This product includes software developed by Stephanie Wehner, QuTech.
# 4. Neither the name of the QuTech organization nor the
#    names of its contributors may be used to endorse or promote products
#    derived from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDER ''AS IS'' AND ANY
# EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER BE LIABLE FOR ANY
# DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import abc
import math
import logging
import warnings
from typing import Any, List
from itertools import count

from cqc.cqcHeader import (
    CQC_VERSION,
    CQC_TP_COMMAND,
    CQC_TP_GET_TIME,
    CQC_TP_HELLO,
    CQC_TP_DONE,
    CQC_TP_NEW_OK,
    CQC_TP_RECV,
    CQC_TP_EPR_OK,
    CQC_TP_MEASOUT,
    CQC_TP_EXPIRE,
    CQC_TP_INF_TIME,
    CQC_ERR_GENERAL,
    CQC_ERR_NOQUBIT,
    CQC_ERR_UNSUPP,
    CQC_ERR_TIMEOUT,
    CQC_ERR_UNKNOWN,
    CQC_CMD_NEW,
    CQC_CMD_SEND,
    CQC_CMD_EPR,
    CQC_CMD_CNOT,
    CQC_CMD_CPHASE,
    CQC_CMD_ROT_X,
    CQC_CMD_ROT_Y,
    CQC_CMD_ROT_Z,
    CQC_CMD_MEASURE,
    CQC_CMD_MEASURE_INPLACE,
    CQC_CMD_RECV,
    CQC_CMD_EPR_RECV,
    CQC_CMD_ALLOCATE,
    CQC_TP_FACTORY,
    Header,
    CQCHeader,
    CQCCmdHeader,
    CQCTypeHeader,
    CQCAssignHeader,
    CQCFactoryHeader,
    CQCRotationHeader,
    CQCXtraQubitHeader,
    CQCCommunicationHeader,
    CQCType,
)
from .util import (
    CQCUnsuppError,
    CQCGeneralError,
    CQCNoQubitError,
    CQCTimeoutError,
    CQCUnknownError,
    ProgressBar,
)
from .qubit import qubit


class CQCHandler(abc.ABC):
    """This class defines the things any CQCHandler must do.

    It is to be subclassed by the various actual classes that handle CQC, such
    as CQCConnection and CQCToFile.
    """

    _appIDs = {}

    def __init__(self, name, app_id=None, pend_messages=False, notify=True):

        self.name = name

        # This flag is used to check if CQCConnection is opened using a 'with' statement.
        # Otherwise an deprecation warning is printed when instantiating qubits.
        self._opened_with_with = False

        # Set an app ID
        self._appID = self._get_new_app_id(app_id)

        self.active_qubits = []

        # This is a sort of global notify
        self.notify = notify

        # All qubits active for this connection
        self.active_qubits = []

        # List of pended header objects waiting to be sent to the backend
        self._pending_headers = []  # ONLY cqc.cqcHeader.Header objects should be in this list

        # Bool that indicates whether we are in a factory and thus should pend commands
        self.pend_messages = pend_messages

        # Keep track of pending messages
        self._pending_headers = []

    @property
    def pend_messages(self):
        return self._pend_messages

    @pend_messages.setter
    def pend_messages(self, value):
        self.set_pending(value)

    def __str__(self):
        return "CQC handler for node '{}'".format(self.name)

    def _get_new_app_id(self, app_id):
        """Finds a new app ID if not specific"""
        name = self.name
        if name not in self._appIDs:
            self._appIDs[name] = []

        # Which appID
        if app_id is None:
            for app_id in count(0):
                if app_id not in self._appIDs[name]:
                    self._appIDs[name].append(app_id)
                    return app_id
        else:
            if app_id in self._appIDs[name]:
                raise ValueError("appID={} is already in use".format(app_id))
            self._appIDs[name].append(app_id)
            return app_id

    def __enter__(self):
        # This flag is used to check if CQCHandler is opened using a 
        # 'with' statement.
        # Otherwise an deprecation warning is printed when instantiating
        # qubits.
        self._opened_with_with = True
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        # All qubits should now be released
        self.close(release_qubits=True)

    @abc.abstractmethod
    def new_qubitID(self, print_cqc=False):
        """Provide new qubit ID.
        
        This method must provide the new qubit ID. This qubit ID could 
        be given by the server you are communicating with, or it might
        simply be a number that increases by one every time the method
        is used. This will depend on the type of CQCHandler being used.
        """
        pass

    @abc.abstractmethod
    def _handle_create_qubits(self, num_qubits, notify):
        """Handles responses after allocating qubits and returns a list of qubits"""
        pass

    @abc.abstractmethod
    def return_meas_outcome(self):
        """Return measurement outcome."""
        pass

    @abc.abstractmethod
    def commit(self, msg):
        """Commit a message. 

        This can mean sending it to the backend or just writing to file.
        """
        pass

    @abc.abstractmethod
    def _handle_factory_response(self, num_iter, response_amount, should_notify=False):
        """Handles the responses from a factory command and returns a list of results"""
        pass

    @abc.abstractmethod
    def get_remote_from_directory_or_address(self, name, **kwargs):
        """Returns the remote address of a given node"""
        pass

    @abc.abstractmethod
    def _handle_epr_response(self, notify):
        """Waits for and handles the response message and returns a qubit object."""
        pass

    @abc.abstractmethod
    def readMessage(self):
        """Receive the whole message from cqc server.

        Returns (CQCHeader,None,None), (CQCHeader,CQCNotifyHeader,None) 
        or (CQCHeader,CQCNotifyHeader,EntInfoHeader) depending on the 
        type of message.
        """
        pass

    def get_appID(self):
        """Returns the application ID."""

        return self._appID

    def commit_command(self, qID, command, notify=1, block=1, action=0, 
                       xtra_qID=0, step=0, remote_appID=0, remote_node=0, 
                       remote_port=0, ref_id=0):
        """Construct and commit command."""

        headers = self.construct_command_headers(
            qID, command, notify=notify, block=block, action=action, 
            xtra_qID=xtra_qID, step=step, remote_appID=remote_appID, 
            remote_node=remote_node, remote_port=remote_port, ref_id=ref_id)
        self.commit_headers(headers)

    def commit_headers(self, headers):
        """Packs a list of headers and commits the message"""
        msg = b''
        for header in headers:
            msg += header.pack()
        self.commit(msg)

    def put_command(self, qID, command, read_notify=True, **kwargs):
        """Puts a new command to be executed.

        If self.pend_messages is set to True, the messages are kept until flushing,
        otherwise they are commited directly.

        Parameters
        ----------
        qID: int
            Id of the qubit to apply the command on
        command: int
            What command to be executed
        read_notify : bool
            Whether to listen to a notify message in this function or if this is handled
            elsewhere (e.g. createEPR)
        """
        headers = self.construct_command_headers(qID=qID, command=command, **kwargs)
        str_of_headers = "".join(["\t{}\n".format(header) for header in headers])
        if self.pend_messages:
            headers = self._update_headers_before_pending(headers)
            logging.debug("App {} pends a command with headers:\n{}".format(
                self.name,
                str_of_headers,
            ))
            self.pend_headers(headers)
        else:
            logging.debug("App {} sends a command with headers:\n{}".format(
                self.name,
                str_of_headers,
            ))
            self.commit_headers(headers)
            if read_notify:
                notify = kwargs.get("notify", True)
                if notify:
                    message = self.readMessage()
                    self._assert_done_message(message)
                    self.print_CQC_msg(message)

    def _update_headers_before_pending(self, headers):
        # Don't include the CQC Headers since this is a sequence
        return headers[1:]

    def _assert_done_message(self, message):
        if message[0].tp != CQC_TP_DONE:
            raise CQCUnsuppError(
                "Unexpected message sent back from the server. Message: {}".format(message[0])
            )

    def pend_headers(self, headers):
        """Pends the given headers"""
        for header in headers:
            self.pend_header(header)
    
    def pend_header(self, header: Header) -> None:
        self._pending_headers.append(header)

    def construct_command(self, qID, command, **kwargs):
        """Construct a commmand and packs it in it's binary form.

        Extra arguments are only used if the command if of a type that 
        needs them.

        - **Arguments**

            :qID:               qubit ID
            :command:           Command to be executed, eg CQC_CMD_H
            :nofify:            Do we wish to be notified when done.
            :block:             Do we want the qubit to be blocked
        """
        headers = self.construct_command_headers(qID, command, **kwargs)
        msg = b''
        for header in headers:
            msg += header.pack()
        return msg

    def construct_command_headers(self, qID, command, **kwargs):
        """Construct a commmand consisting of a list of header objects.

        Extra arguments are only used if the command if of a type that 
        needs them.

        - **Arguments**

            :qID:               qubit ID
            :command:           Command to be executed, eg CQC_CMD_H
            :nofify:            Do we wish to be notified when done.
            :block:             Do we want the qubit to be blocked
        """
        # Construct extra header if needed.
        xtra_hdr = None
        if command == CQC_CMD_SEND or command == CQC_CMD_EPR:
            xtra_hdr = CQCCommunicationHeader()
            remote_appID = kwargs.get("remote_appID", 0)
            remote_node = kwargs.get("remote_node", 0)
            remote_port = kwargs.get("remote_port", 0)
            xtra_hdr.setVals(remote_appID, remote_node, remote_port)
        elif command == CQC_CMD_CNOT or command == CQC_CMD_CPHASE:
            xtra_hdr = CQCXtraQubitHeader()
            xtra_qID = kwargs.get("xtra_qID", 0)
            xtra_hdr.setVals(xtra_qID)
        elif (command == CQC_CMD_ROT_X or command == CQC_CMD_ROT_Y 
              or command == CQC_CMD_ROT_Z):
            xtra_hdr = CQCRotationHeader()
            step = kwargs.get("step", 0)
            xtra_hdr.setVals(step)
        elif command == CQC_CMD_MEASURE or command == CQC_CMD_MEASURE_INPLACE:
            xtra_hdr = CQCAssignHeader()
            ref_id = kwargs.get("ref_id", 0)
            xtra_hdr.setVals(ref_id)

        # If xtra_hdr is None, we don't need an extra message.
        if xtra_hdr is None:
            header_length = CQCCmdHeader.HDR_LENGTH
        else:
            header_length = CQCCmdHeader.HDR_LENGTH + xtra_hdr.HDR_LENGTH

        # Construct Header
        hdr = CQCHeader()
        hdr.setVals(CQC_VERSION, CQC_TP_COMMAND, self._appID, header_length)

        # Construct Command
        cmd_hdr = CQCCmdHeader()
        notify = int(kwargs.get("notify", True))
        block = int(kwargs.get("block", True))
        action = int(kwargs.get("action", False))
        cmd_hdr.setVals(qID, command, notify, block, action)

        headers = [hdr, cmd_hdr]
        if xtra_hdr is not None:
            headers.append(xtra_hdr)

        return headers

    def construct_simple(self, tp):
        """Construct simple message.
        
        For example a HELLO message if tp=CQC_TP_HELLO.
        """
        hdr = CQCHeader()
        hdr.setVals(CQC_VERSION, tp, self._appID, 0)
        msg = hdr.pack()
        return msg

    def sendSimple(self, tp):
        """Construct and commit simple message."""
        msg = self.construct_simple(tp)
        self.commit(msg)

    def close(self, release_qubits=True):
        """Handle exiting of context."""

        if release_qubits:
            for q in list(self.active_qubits):
                q.release()

        # Flush all remaining commands and the releases
        self.flush()

        self._pop_app_id()

    def _pop_app_id(self):
        """
        Removes the used appID from the list.
        """
        try:
            self._appIDs[self.name].remove(self._appID)
        except ValueError:
            pass  # Already removed

    def create_qubits(self, num_qubits, block=True, notify=True):
        """Requests the backend to reserve some qubits

        :param num_qubits: The amount of qubits to reserve
        :return: A list of qubits
        :param notify:     Do we wish to be notified when done.
        :param block:         Do we want the qubit to be blocked
        """
        notify = self.notify and notify

        # TODO how to handle pending headers?
        headers = self.construct_command_headers(
            qID=num_qubits,
            command=CQC_CMD_ALLOCATE,
            notify=notify,
            block=block,
        )
        self.commit_headers(headers)

        qubits = self._handle_create_qubits(num_qubits=num_qubits, notify=notify)

        return qubits

    def sendGetTime(self, qID, notify=1, block=1, action=0):
        """Sends get-time message

        - **Arguments**

            :qID:         qubit ID
            :command:     Command to be executed, eg CQC_CMD_H
            :notify:     Do we wish to be notified when done.
            :block:         Do we want the qubit to be blocked
            :action:     Are there more commands to be executed
        """
        # Send Header
        hdr = CQCHeader()
        hdr.setVals(CQC_VERSION, CQC_TP_GET_TIME, self._appID, CQCCmdHeader.HDR_LENGTH)
        msg = hdr.pack()
        self.commit(msg)

        # Send Command
        cmd_hdr = CQCCmdHeader()
        cmd_hdr.setVals(qID, 0, notify, block, action)
        cmd_msg = cmd_hdr.pack()
        self.commit(cmd_msg)

    def allocate_qubits(self, nb_of_qubits: int) -> List['qubit']:
        """
        Creates (i.e. allocates) multiple qubits, and returns a list with qubit objects.

        :nb_of_qubits: The amount of qubits to be created.
        """
        warnings.warn("allocate_qubits is deprecated, use create_qubits instead",
                      DeprecationWarning)

        return self.create_qubits(nb_of_qubits)

    def createEPR(self, name, remote_appID=0, notify=True, block=True, **kwargs):
        """Creates epr with other host in the network.

        - **Arguments**

            :name:         Name of the node as specified in the cqc network config file.
            :remote_appID:     The app ID of the application running on the receiving node.
            :nofify:     Do we wish to be notified when done.
            :block:         Do we want the qubit to be blocked
        """
        remote_ip, remote_port = self.get_remote_from_directory_or_address(name, **kwargs)

        # print info
        logging.debug(
            "App {} puts message: 'Create EPR-pair with {} and appID {}'".format(self.name, name, remote_appID)
        )
        notify = self.notify and notify
        self.put_command(
            0,
            CQC_CMD_EPR,
            read_notify=False,
            notify=notify,
            block=block,
            remote_appID=remote_appID,
            remote_node=remote_ip,
            remote_port=remote_port,
        )
        if not self.pend_messages:
            q = self._handle_epr_response(notify=notify)
            return q

    def recvEPR(self, notify=True, block=True):
        """Receives a qubit from an EPR-pair generated with another node.

        - **Arguments**

            :nofify:     Do we wish to be notified when done.
            :block:         Do we want the qubit to be blocked
        """
        # print info
        logging.debug("App {} puts message: 'Receive half of EPR'".format(self.name))
        notify = self.notify and notify
        self.put_command(
            qID=0,
            command=CQC_CMD_EPR_RECV,
            read_notify=False,
            notify=notify,
            block=block,
        )

        if not self.pend_messages:
            q = self._handle_epr_response(notify=notify)
            return q

    def sendQubit(self, q, name, remote_appID=0, notify=True, block=True, **kwargs):
        """Sends qubit to another node in the cqc network. 
        
        If this node is not in the network an error is raised.

        - **Arguments**

            :q:         The qubit to send.
            :Name:         Name of the node as specified in the cqc network config file.
            :remote_appID:     The app ID of the application running on the receiving node.
            :nofify:     Do we wish to be notified when done.
            :block:         Do we want the qubit to be blocked
        """
        remote_ip, remote_port = self.get_remote_from_directory_or_address(name, **kwargs)

        # print info
        logging.debug(
            "App {} puts message: 'Send qubit with ID {} to {} and appID {}'".format(
                self.name, q._qID, name, remote_appID
            )
        )
        notify = self.notify and notify
        self.put_command(
            qID=q._qID,
            command=CQC_CMD_SEND,
            notify=notify, 
            block=block,
            remote_appID=remote_appID, 
            remote_node=remote_ip,
            remote_port=remote_port,
        )
        # Deactivate qubit
        # TODO should this be done if pending messages?
        q._set_active(False)

    def recvQubit(self, notify=True, block=True):
        """Receives a qubit.

        - **Arguments**

            :q:         The qubit to send.
            :Name:         Name of the node as specified in the cqc network config file.
            :remote_appID:     The app ID of the application running on the receiving node.
            :nofify:     Do we wish to be notified when done.
            :block:         Do we want the qubit to be blocked
        """

        # print info
        logging.debug("App {} puts message: 'Receive qubit'".format(self.name))
        notify = self.notify and notify
        self.put_command(0, CQC_CMD_RECV, read_notify=False, notify=notify, block=block)
        if not self.pend_messages:
            # Get qubit id
            q_id = self.new_qubitID(print_cqc=True)

            # initialize the qubit
            q = qubit(self, createNew=False)
            q._qID = q_id

            # Activate and return qubit
            q._set_active(True)

            # Read the notify message
            if notify:
                message = self.readMessage()
                self._assert_done_message(message)
                self.print_CQC_msg(message)

            return q

    def flush(self, do_sequence=False):
        """Flush all pending messages to the backend.
        
        :param do_sequence: boolean to indicate if you want to send the pending messages as a sequence
        :return: A list of things that are sent back from the server. Can be qubits, or outcomes
        """
        return self.flush_factory(1, do_sequence)

    def flush_factory(self, num_iter, do_sequence=False, block_factory=False):
        """
        Flushes the current pending sequence in a factory. It is performed multiple times
        :param num_iter: The amount of times the current pending sequence is performed
        :return: A list of outcomes/qubits that are produced by the commands
        """
        if len(self._pending_headers) == 0:
            return []

        # Initialize should_notify to False
        should_notify = False

        # Store how many of the headers we send will get a response message from the backend
        response_amount = 0

        # Loop over the pending_headers to determine the total length and set should_notify
        for header in self._pending_headers:

            # Check if the current header is a Command header. It can also be a sub header
            if isinstance(header, CQCCmdHeader):
                # set should_notify to True if at least one of all command headers has notify to True
                should_notify = should_notify or header.notify
                
                # Remember this header if we expect a return messge
                if self.shouldReturn(header.instr):
                    response_amount += 1

        # Determine the CQC Header type
        if num_iter == 1:
            cqc_type = CQC_TP_COMMAND
        else:
            # Build and insert the Factory header
            cqc_type = CQC_TP_FACTORY
            factory_header = CQCFactoryHeader()
            factory_header.setVals(num_iter, should_notify, block_factory)
            # Insert the factory header at the front
            self._pending_headers.insert(0, factory_header)
            
        # Insert the cqc header
        self.insert_cqc_header(cqc_type)
        
        # Send all pending headers
        self.send_pending_headers()

        # Reset _pending_headers to an empty list after all headers are sent
        self.reset_pending_headers()

        # Read out any returned messages from the backend
        res = self._handle_factory_response(num_iter, response_amount, should_notify=should_notify)
        
        # Return information that the backend returned
        return res

    def send_pending_headers(self) -> List[Any]:
        """
        Sends all pending headers.
        After sending, self._pending_headers is emptied.
        """

        # Send all pending headers
        to_log = "App {} sends a message with the following headers:\n".format(self.name)
        msg = b''
        for header in self._pending_headers:
            to_log += "\t{}\n".format(header)
            msg += header.pack()
        logging.debug(to_log[:-1])
        self.commit(msg)

    def reset_pending_headers(self):
        """Sets the list of pending headers to empty """
        self._pending_headers = []

    def set_pending(self, pend_messages):
        """Set the pend_messages flag.

        If true, flush() has to be called to send all self._pending_headers in sequence to the backend
        If false, all commands are directly send to the back_end
        :param pend_messages: Boolean to indicate if messages should pend or not
        """
        # Check if the list is not empty, give a warning if it isn't
        if self._pending_headers:
            logging.warning("List of pending headers is not empty, flushing them")
            self.flush()
        self._pend_messages = pend_messages

    def insert_cqc_header(self, cqc_type: CQCType, version=CQC_VERSION) -> None:
        """
        Inserts a CQC Header at index 0 of self._pending_headers.
        Invoke this method *after* all other headers are pended, so that the correct message length is calculated.
        """

        # Count the total message length
        message_length = 0
        for header in self._pending_headers:
            message_length += header.HDR_LENGTH

        # Build the CQC Header
        cqc_header = CQCHeader()
        cqc_header.setVals(CQC_VERSION, cqc_type, self._appID, message_length)

        # Insert CQC Header at the front
        self._pending_headers.insert(0, cqc_header)

    def _pend_type_header(self, cqc_type: CQCType, length: int) -> None:
        """
        Creates a CQCTypeHeader and pends it.
        """
        header = CQCTypeHeader()
        header.setVals(cqc_type, length)
        self.pend_header(header)

    def tomography(self, preparation, iterations, progress=True):
        """
        Does a tomography on the output from the preparation specified.
        The frequencies from X, Y and Z measurements are returned as a tuple (f_X,f_Y,f_Z).

        - **Arguments**

            :preparation:     A function that takes a CQCConnection as input and prepares a qubit and returns this
            :iterations:     Number of measurements in each basis.
            :progress_bar:     Displays a progress bar
        """
        accum_outcomes = [0, 0, 0]
        if progress:
            bar = ProgressBar(3 * iterations)

            # Measure in X
        for _ in range(iterations):
            # Progress bar
            if progress:
                bar.increase()

            # prepare and measure
            q = preparation(self)
            q.H()
            m = q.measure()
            accum_outcomes[0] += m

            # Measure in Y
        for _ in range(iterations):
            # Progress bar
            if progress:
                bar.increase()

            # prepare and measure
            q = preparation(self)
            q.K()
            m = q.measure()
            accum_outcomes[1] += m

            # Measure in Z
        for _ in range(iterations):
            # Progress bar
            if progress:
                bar.increase()

            # prepare and measure
            q = preparation(self)
            m = q.measure()
            accum_outcomes[2] += m

        if progress:
            bar.close()
            del bar

        freqs = map(lambda x: x / iterations, accum_outcomes)
        return list(freqs)

    def test_preparation(self, preparation, exp_values, conf=2, iterations=100, progress=True):
        """Test the preparation of a qubit.
        Returns True if the expected values are inside the confidence interval produced from the data received from
        the tomography function

        - **Arguments**

            :preparation:     A function that takes a CQCConnection as input and prepares a qubit and returns this
            :exp_values:     The expected values for measurements in the X, Y and Z basis.
            :conf:         Determines the confidence region (+/- conf/sqrt(iterations) )
            :iterations:     Number of measurements in each basis.
            :progress_bar:     Displays a progress bar
        """
        epsilon = conf / math.sqrt(iterations)

        freqs = self.tomography(preparation, iterations, progress=progress)
        for i in range(3):
            if abs(freqs[i] - exp_values[i]) > epsilon:
                print(freqs, exp_values, epsilon)
                return False
        return True

    def print_CQC_msg(self, message):
        """
        Prints messsage returned by the readMessage method of CQCConnection.
        """
        # First check if there was an error
        self.check_error(message[0])

        hdr = message[0]
        otherHdr = message[1]
        entInfoHdr = message[2]

        if hdr.tp == CQC_TP_HELLO:
            logging.debug("CQC tells App {}: 'HELLO'".format(self.name))
        elif hdr.tp == CQC_TP_EXPIRE:
            logging.debug("CQC tells App {}: 'Qubit with ID {} has expired'".format(self.name, otherHdr.qubit_id))
        elif hdr.tp == CQC_TP_DONE:
            logging.debug("CQC tells App {}: 'Done with command'".format(self.name))
        elif hdr.tp == CQC_TP_RECV:
            logging.debug("CQC tells App {}: 'Received qubit with ID {}'".format(self.name, otherHdr.qubit_id))
        elif hdr.tp == CQC_TP_EPR_OK:

            # Lookup host name
            remote_node = entInfoHdr.node_B
            remote_port = entInfoHdr.port_B
            remote_name = None
            try:
                for node in self._cqcNet.hostDict.values():
                    if (node.ip == remote_node) and (node.port == remote_port):
                        remote_name = node.name
                        break
                if remote_name is None:
                    raise RuntimeError("Remote node ({},{}) is not in config-file.".format(remote_node, remote_port))
            except AttributeError:
                remote_name = "({}, {})".format(remote_node, remote_port)

            logging.debug(
                "CQC tells App {}: 'EPR created with node {}, using qubit with ID {}'".format(
                    self.name, remote_name, otherHdr.qubit_id
                )
            )
        elif hdr.tp == CQC_TP_MEASOUT:
            logging.debug("CQC tells App {}: 'Measurement outcome is {}'".format(self.name, otherHdr.outcome))
        elif hdr.tp == CQC_TP_INF_TIME:
            logging.debug("CQC tells App {}: 'Timestamp is {}'".format(self.name, otherHdr.datetime))

    def parse_CQC_msg(self, message, q=None, is_factory=False):
        """
        parses the cqc message and returns the relevant value of that measure
        (qubit, measurement outcome)

        :param message: str
            the cqc message to be parsed
        :param q: :obj:`cqc.pythonLib.qubit`
            the qubit object we should save the qubit to
        :param is_factory: bool
            whether the returned message came from a factory. If so, do not change the qubit, but create a new one
        :return: the result of the message (either a qubit, or a measurement outcome. Otherwise None
        """
        hdr = message[0]
        otherHdr = message[1]
        if len(message) < 3:
            entInfoHdr = None
        else:
            entInfoHdr = message[2]

        if hdr.tp in {CQC_TP_RECV, CQC_TP_NEW_OK, CQC_TP_EPR_OK}:
            if is_factory:
                q._set_active(False)  # Set qubit to inactive so it can't be used anymore
                q = qubit(self, createNew=False)
            if q is None:
                q = qubit(self, createNew=False)
            q._qID = otherHdr.qubit_id
            q._set_entanglement_info(entInfoHdr)
            q._set_active(True)
            return q
        if hdr.tp == CQC_TP_MEASOUT:
            return otherHdr.outcome
        if hdr.tp == CQC_TP_INF_TIME:
            return otherHdr.datetime

    def check_error(self, hdr):
        """Checks if there is an error returned."""

        self._errorHandler(hdr.tp)

    def _errorHandler(self, cqc_err):
        """Raises an error if there is an error-message."""

        if cqc_err == CQC_ERR_GENERAL:
            raise CQCGeneralError("General error")
        if cqc_err == CQC_ERR_NOQUBIT:
            raise CQCNoQubitError("No more qubits available")
        if cqc_err == CQC_ERR_UNSUPP:
            raise CQCUnsuppError("Sequence not supported")
        if cqc_err == CQC_ERR_TIMEOUT:
            raise CQCTimeoutError("Timeout")
        if cqc_err == CQC_ERR_UNKNOWN:
            raise CQCUnknownError("Unknown qubit ID")

    @staticmethod
    def shouldReturn(command):
        return command in {
            CQC_CMD_NEW,
            CQC_CMD_MEASURE,
            CQC_CMD_MEASURE_INPLACE,
            CQC_CMD_RECV,
            CQC_CMD_EPR_RECV,
            CQC_CMD_EPR,
        }

    @staticmethod
    def hasXtraHeader(command):
        return command in {
            CQC_CMD_CNOT,
            CQC_CMD_SEND,
            CQC_CMD_EPR,
            CQC_CMD_ROT_X,
            CQC_CMD_ROT_Y,
            CQC_CMD_ROT_Z,
            CQC_CMD_CPHASE,
        }



#
# Copyright (c) 2017, Stephanie Wehner and Axel Dahlberg
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
# 1. Redistributions of source code must retain the above copyright
#    notice, this list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright
#    notice, this list of conditions and the following disclaimer in the
#    documentation and/or other materials provided with the distribution.
# 3. All advertising materials mentioning features or use of this software
#    must display the following acknowledgement:
#    This product includes software developed by Stephanie Wehner, QuTech.
# 4. Neither the name of the QuTech organization nor the
#    names of its contributors may be used to endorse or promote products
#    derived from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDER ''AS IS'' AND ANY
# EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER BE LIABLE FOR ANY
# DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.


import os
import sys


class ProgressBar:
    def __init__(self, maxitr):
        self.maxitr = maxitr
        self.itr = 0
        try:
            self.cols = os.get_terminal_size().columns
        except (OSError, AttributeError):
            self.cols = 60
        print("")
        self.update()

    def increase(self):
        self.itr += 1
        self.update()

    def update(self):
        cols = self.cols - 8
        assert self.itr <= self.maxitr
        ratio = float(self.itr) / self.maxitr
        procent = int(ratio * 100)
        progress = "=" * int(cols * ratio)
        sys.stdout.write("\r")
        sys.stdout.write("[%*s] %d%%" % (-cols, progress, procent))
        sys.stdout.flush()
        pass

    def close(self):
        print("")


class CQCGeneralError(Exception):
    pass


class CQCNoQubitError(CQCGeneralError):
    pass


class CQCUnsuppError(CQCGeneralError):
    pass


class CQCTimeoutError(CQCGeneralError):
    pass


class CQCInuseError(CQCGeneralError):
    pass


class CQCUnknownError(CQCGeneralError):
    pass


class QubitNotActiveError(CQCGeneralError):
    pass



#
# Copyright (c) 2017, Stephanie Wehner and Axel Dahlberg
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
# 1. Redistributions of source code must retain the above copyright
#    notice, this list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright
#    notice, this list of conditions and the following disclaimer in the
#    documentation and/or other materials provided with the distribution.
# 3. All advertising materials mentioning features or use of this software
#    must display the following acknowledgement:
#    This product includes software developed by Stephanie Wehner, QuTech.
# 4. Neither the name of the QuTech organization nor the
#    names of its contributors may be used to endorse or promote products
#    derived from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDER ''AS IS'' AND ANY
# EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER BE LIABLE FOR ANY
# DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES
# LOSS OF USE, DATA, OR PROFITS OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

from cqc.cqcHeader import (
    CQCCmdHeader,
    CQCHeader,
    CQCTypeHeader,
    CQCAssignHeader,
    CQCFactoryHeader,
    CQCIfHeader,
    CQCXtraQubitHeader,
    CQCType,
    CQC_CMD_CNOT,
    CQC_CMD_CPHASE,
    CQC_CMD_MEASURE,
    CQC_CMD_MEASURE_INPLACE,
)


def parse_cqc_message(msg):
    # TODO finish parser
    hdr_map = {
        CQCType.HELLO: None,
        CQCType.COMMAND: CQCCmdHeader,
        CQCType.MIX: CQCTypeHeader,
        CQCType.IF: CQCIfHeader,
        CQCType.FACTORY: CQCFactoryHeader,
    }
    headers = []
    next_hdr = CQCHeader
    bits_to_next_first_hdr = 0
    is_mix = False
    while len(msg) > 0:
        hdr = extract_header(msg, next_hdr)
        headers.append(hdr)
        next_hdr = None
        msg = msg[hdr.HDR_LENGTH:]
        if isinstance(hdr, CQCHeader):
            bits_to_next_first_hdr = hdr.length
            next_hdr = hdr_map[hdr.tp]
            is_mix = hdr.tp == CQCType.MIX
        else:
            if isinstance(hdr, CQCCmdHeader):
                if hdr.instr in [CQC_CMD_MEASURE, CQC_CMD_MEASURE_INPLACE]:
                    next_hdr = CQCAssignHeader
                if hdr.instr in [CQC_CMD_CNOT, CQC_CMD_CPHASE]:
                    next_hdr = CQCXtraQubitHeader
            elif isinstance(hdr, CQCTypeHeader):
                next_hdr = hdr_map[hdr.type]
            elif isinstance(hdr, CQCAssignHeader):
                pass
            elif isinstance(hdr, CQCIfHeader):
                pass
            elif isinstance(hdr, CQCXtraQubitHeader):
                pass
            elif isinstance(hdr, CQCFactoryHeader):
                next_hdr = CQCCmdHeader
            else:
                raise NotImplementedError("for {}".format(hdr))
            bits_to_next_first_hdr -= hdr.HDR_LENGTH
        if bits_to_next_first_hdr <= 0:
            next_hdr = CQCHeader
        if next_hdr is None:
            if is_mix:
                next_hdr = CQCTypeHeader
            else:
                next_hdr = CQCCmdHeader

    return headers


def extract_header(msg, hdr_class):
    hdr = hdr_class(msg[:hdr_class.HDR_LENGTH])
    return hdr



#
# Copyright (c) 2017, Stephanie Wehner and Axel Dahlberg
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
# 1. Redistributions of source code must retain the above copyright
#    notice, this list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright
#    notice, this list of conditions and the following disclaimer in the
#    documentation and/or other materials provided with the distribution.
# 3. All advertising materials mentioning features or use of this software
#    must display the following acknowledgement:
#    This product includes software developed by Stephanie Wehner, QuTech.
# 4. Neither the name of the QuTech organization nor the
#    names of its contributors may be used to endorse or promote products
#    derived from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDER ''AS IS'' AND ANY
# EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER BE LIABLE FOR ANY
# DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

from typing import Union
from anytree import NodeMixin

from cqc.cqcHeader import (
    CQC_CMD_MEASURE,
    CQC_CMD_MEASURE_INPLACE,
    CQCIfHeader,
    CQCTypeHeader,
    CQCFactoryHeader,
    CQCType,
    CQCLogicalOperator,
)
from .util import CQCGeneralError, QubitNotActiveError
from .qubit import qubit
from .cqc_connection import CQCConnection


class CQCVariable:
    """
    Instances of this class are returned by measure command, if executed inside a CQCMix context.
    A CQCVariable holds a reference ID with which one can refer to the outcome of the measurement.
    """
    _next_ref_id = 0
    
    def __init__(self):
        """
        Increments the reference ID, and assigns the new unique reference ID to this CQCVariable.
        This system ensures no two CQCVariable instances have the same reference ID.
        """
        self._ref_id = CQCVariable._next_ref_id
        CQCVariable._next_ref_id += 1

    # make ref_id a read-only variable
    @property
    def ref_id(self):
        """
        Get the refernce ID of this CQCVariable. This is a read-only property.
        """
        return self._ref_id

    # override the == operator
    # other can be a CQCVariable or int
    def __eq__(self, other: Union['CQCVariable', int]):
        return _LogicalFunction(self, CQCLogicalOperator.EQ, other)
    
    # override the != operator
    def __ne__(self, other: Union['CQCVariable', int]):
        return _LogicalFunction(self, CQCLogicalOperator.NEQ, other)


class _LogicalFunction:
    """
    Private helper class. This class should never be used outside this pythonLib.
    """

    def __init__(
        self, 
        operand_one: CQCVariable, 
        operator: CQCLogicalOperator, 
        operand_two: Union[CQCVariable, int]
    ):
        """
        Stores all information necessary to create a logical comparison

        - **Arguments**

            :operand_one:   The CQCVariable that stores the measurement outcome that must be compared
            :operator:      One of the CQCLogicalOperator types that CQC supports. 
                            At present, equality and inequality are supported.
            :operand_two:   Either a CQCVariable or an integer. 
                            If a CQCVariable, then the value behind this variable will be compared to operand_one. 
                            If an integer, then the value behind operand_one will be compared to this integer.
        """

        self.operand_one = operand_one
        self.operator = operator
        self.operand_two = operand_two

    def get_negation(self) -> '_LogicalFunction':
        return _LogicalFunction(self.operand_one, CQCLogicalOperator.opposite_of(self.operator), self.operand_two)

    def get_CQCIfHeader(self) -> CQCIfHeader:
        """
        Builds the If header corresponding to this logical function.
        """

        if isinstance(self.operand_two, int):
            type_of_operand_two = CQCIfHeader.TYPE_VALUE
            operand_two = self.operand_two
        else:
            type_of_operand_two = CQCIfHeader.TYPE_REF_ID
            operand_two = self.operand_two._ref_id

        header = CQCIfHeader()
        header.setVals(
            self.operand_one.ref_id,
            self.operator,
            type_of_operand_two,
            operand_two,
            length=0
        )
        return header


class CQCMixConnection(CQCConnection):
    """Subclass of CQCconnection to be used with CQCMix"""
    def __init__(self, name, socket_address=None, appID=None, pend_messages=False,
                 retry_connection=True, conn_retry_time=0.1, log_level=None, backend=None,
                 use_classical_communication=True, network_name=None):
        super().__init__(
            name=name,
            socket_address=socket_address,
            appID=appID,
            pend_messages=pend_messages,
            retry_connection=retry_connection,
            conn_retry_time=conn_retry_time,
            log_level=log_level,
            backend=backend,
            use_classical_communication=use_classical_communication,
            network_name=network_name,
        )

        # Variable of type NodeMixin. This variable is used in CQCMix types to create a
        # scoping mechanism.
        self.current_scope = None

        self._inside_cqc_mix = False

    def _update_headers_before_pending(self, headers):
        # Insert type headers if in cqc mix
        if self._inside_cqc_mix:
            length = sum([hdr.HDR_LENGTH for hdr in headers[1:]])
            tp_header = CQCTypeHeader()
            tp_header.setVals(CQCType.COMMAND, length)
            return [tp_header] + headers[1:]
        else:
            return headers[1:]

    def _enter_mix(self):
        # Set the _inside_cqc_mix bool to True on the connection
        self._inside_cqc_mix = True

        self.pend_messages = True


class mix_qubit(qubit):
    def __init__(self, cqc: CQCMixConnection, notify=True, block=True, createNew=True, q_id=None, entInfo=None):
        # This stores the scope (type NodeMixin) in which this qubit was deactivated
        # If the qubit has not yet been deactivated, this is set to None
        self.scope_of_deactivation = None

        super().__init__(
            cqc=cqc,
            notify=notify,
            block=block,
            createNew=createNew,
            q_id=q_id,
            entInfo=entInfo,
        )

    def check_active(self):
        """
        Checks if the qubit is active
        """
        if not self._active:

            # This conditional checks whether it is certain that the qubit is inactive at this 
            # point in the code. If such is the case, an error is raised. 
            # At this point, it is certain that self_active is False. However, this does not necessarily
            # mean that the qubit is inactive due to the possibility to write cqc_if blocks.
            # There are four options:
            # 1) Control is currently not inside a CQCMix. In that case, the qubit is inactive.
            # 2) The qubit was deactivated in the current scope. The qubit therefore is inactive.
            # 3) The qubit was deactivated in an ancestor scope. The qubit therefore is inactive.
            # 4) The qubit was deactivated in a descendent scope.  The qubit is therefore inactive. 
            # The only possible way self_active can be False but the qubit is in fact active, is
            # if the qubit was deactivated in a sibling scope, such as the sibling if-block of an else-block.
            if (
                not self._cqc._inside_cqc_mix
                or self.scope_of_deactivation == self._cqc.current_scope
                or self.scope_of_deactivation in self._cqc.current_scope.ancestors
                or self.scope_of_deactivation in self._cqc.current_scope.descendants
            ):

                raise QubitNotActiveError(
                    "Qubit is not active. Possible causes:\n"
                    "- Qubit is sent to another node\n"
                    "- Qubit is measured (with inplace=False)\n"
                    "- Qubit is released\n"
                    "- Qubit is not received\n"
                    "- Qubit is used and created in the same factory\n"
                    "- Qubit is measured (with inplace=False) inside a cqc_if block earlier in the code\n"
                )

    def _set_active(self, be_active):

        # Set the scope of deactivation to the current scope, if inside a CQCMix.
        if not be_active and self._cqc._inside_cqc_mix:
            self.scope_of_deactivation = self._cqc.current_scope

        super()._set_active(be_active)

    def measure(self, inplace=False, block=True):
        """
        Measures the qubit in the standard basis and returns the measurement outcome.
        If now MEASOUT message is received, None is returned.
        If inplace=False, the measurement is destructive and the qubit is removed from memory.
        If inplace=True, the qubit is left in the post-measurement state.

        - **Arguments**

            :inplace:     If false, measure destructively.
            :block:         Do we want the qubit to be blocked
        """
        # check if qubit is active
        self.check_active()

        if inplace:
            command = CQC_CMD_MEASURE_INPLACE
        else:
            command = CQC_CMD_MEASURE
            # Set qubit to non active so the user can receive helpful errors during compile time 
            # if this qubit is used after this measurement
            self._set_active(False)

        if self._cqc.pend_messages:
            cqc_variable = CQCVariable()
            ref_id = cqc_variable.ref_id
        else:
            ref_id = 0

        self._cqc.put_command(
            qID=self._qID,
            command=command,
            notify=False,
            block=block,
            ref_id=ref_id,
        )

        if self._cqc.pend_messages:
            return cqc_variable
        else:
            return self._cqc.return_meas_outcome()


class CQCMix(NodeMixin):
    """
    This Python Context Manager Type can be used to create CQC programs that consist of more than a single type.
    Hence the name CQC Mix. Programs of this type can consist of any number and mix of the other CQC types. 
    """

    def __init__(self, cqc_connection: CQCMixConnection):
        """
        Initializes the Mix context.

        - **Arguments**

            :cqc_connection:    The CQCConnection to which this CQC Program must be sent.
        """
        if not isinstance(cqc_connection, CQCMixConnection):
            raise TypeError("To use CQCMix the connection needs to be of type CQCMixConnection, "
                            "not {}".format(type(cqc_connection)))

        self._conn = cqc_connection

        # Set the current scope to self
        self._conn.current_scope = self

    def __enter__(self):
        # Update the connection to be compatible with mixing
        self._conn._enter_mix()

        # Return self so that this instance is bound to the variable after "as", i.e.: "with CQCMix() as pgrm"
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        
        # Only do these things if there was no exception.
        if exc_type is None:
            # Build and insert the CQC Header
            self._conn.insert_cqc_header(CQCType.MIX)

            # Send this program to the backend
            self._conn.send_pending_headers()
            self._conn.reset_pending_headers()

            # We expect one message back, which can be an error or TP_DONE
            # This also blocks the program until we have received a message from the backend, 
            # which is important because it avoids that we send more messages before the backend is finished.
            message = self._conn.readMessage()

            # Check if it is an error and assume it is a TP_DONE if it is not an error
            self._conn.check_error(message[0])

            # We are no longer in a TP_MIX
            self._conn._inside_cqc_mix = False

            self._conn.pend_messages = False

            # Set the current scope to None, since we exit the CQCMix context 
            # current_scope is only used inside CQCMix contexts
            self._conn.current_scope = None

    def cqc_if(self, logical_function: _LogicalFunction):
        """
        Open a Python Context Manager Type to start an if-statement block.

        - **Arguments**

            :logical_function:      A _LogicalFunction instance. Never instantiate this explicitely; instead
                                    use the following: CQCVariable == 1 OR CQCVariable == CQCVariable. 
                                    CQCVariable can be any instance that you want to test to a value, or to another  
                                    CQCVariable. The operator can be == or !=. 
                                    The value can be any integer (though only 1 and 0 make sense).
                                
        """
        return _CQCConditional(self._conn, False, logical_function)

    def cqc_else(self):
        """
        Open a Python Context Manager Type to start an else-statement block.
        This will be  an else-block of the last closed cqc_if-block.                    
        """
        # Find out to which if this else belongs
        return _CQCConditional(self._conn, True)

    def loop(self, times: int):
        """
        Open a Python Context Manager Type to start a factory (i.e. repeated sequence of commands).

        - **Arguments**

            :times:     The number of times the commands inside body of this context should be repeated.
                                
        """
        return _CQCFactory(self._conn, times)


class _CQCFactory:
    """
    Private class to create factories inside CQCMix contexts. Never explicitely instantiate this class outside 
    the source code of this library.
    Instead, use CQCMix.loop(x), where x is the amount of times to repeat.
    """

    def __init__(self, cqc_connection, repetition_amount: int):
        self._conn = cqc_connection
        self._repetition_amount = repetition_amount

    def __enter__(self):

        # Inside a TP_FACTORY, we don't want CQCType headers before every instruction.
        # Therefore, we set this bool to False
        self._conn._inside_cqc_mix = False

        # Create the CQC Type header, and store it so that we can modify its length at __exit__
        self.type_header = CQCTypeHeader()
        self.type_header.setVals(CQCType.FACTORY, length=0)

        # Build the Factory header
        factory_header = CQCFactoryHeader()
        factory_header.setVals(self._repetition_amount)

        # Pend the headers
        self._conn.pend_header(self.type_header)
        self._conn.pend_header(factory_header)

    def __exit__(self, exc_type, exc_val, exc_tb):

        # Outside a TP_FACTORY, we want CQCType headers before every instruction.
        # Therefore, we set this bool to True
        self._conn._inside_cqc_mix = True

        # Calculate the length of the body of the factory
        # Loop in reverse through all pending_headers to calculate the length of all headers
        index = len(self._conn._pending_headers) - 1
        body_length = 0
        while self._conn._pending_headers[index] is not self.type_header:
            body_length += self._conn._pending_headers[index].HDR_LENGTH
            index -= 1
        
        # Set the correct length
        self.type_header.length = body_length


class _CQCConditional(NodeMixin):
    """
    Private helper class. Never explicitely instantiate this class outside the source code of this library.
    This Context Manager class is instantiated by CQCMix.cqc_if() and CQCMix.cqc_else(). Its 
    function is to build and pend CQC If headers.
    """

    # This private class variable holds the last _CQCConditional that 
    # functioned as an IF (as opposed to an ELSE) on which __exit__ is invoked. 
    # In other words, it is the last closed IF statement. 
    # This is important so that ELSE statements can find out to which IF statement they belong.
    # If this variable is None, then there either has not been aan IF statement yet, or the last 
    # _CQCConditional was an ELSE.
    _last_closed_conditional = None

    def __init__(self, cqc_connection, is_else: bool, logical_function: _LogicalFunction = None):
        self._conn = cqc_connection
        self.is_else = is_else

        if is_else:
            # If _last_closed_conditional is None, then there either has not been aan IF statement yet, or the last 
            # _CQCConditional was an ELSE.
            if _CQCConditional._last_closed_conditional is None:
                raise CQCGeneralError('Cannot use an ELSE if there is no IF directly before it.')
            else:
                # Get the negation of the logical function of the IF, 
                # which will be the logical function for this ELSE statement
                logical_function = _CQCConditional._last_closed_conditional._logical_function.get_negation()
            
        self._logical_function = logical_function

    def __enter__(self):
        # Pend CQC Type header
        self._conn._pend_type_header(CQCType.IF, CQCIfHeader.HDR_LENGTH)

        # Build the IF header, and store it so we can modify its length at __exit__
        self.header = self._logical_function.get_CQCIfHeader()

        # Pend the IF header
        self._conn.pend_header(self.header)

        # Register the parent scope, and set the current scope to self
        self.parent = self._conn.current_scope
        self._conn.current_scope = self

    def __exit__(self, exc_type, exc_val, exc_tb):

        # Set _last_closed_conditional to the correct value
        if (self.is_else):
            _CQCConditional._last_closed_conditional = None
        else:
            _CQCConditional._last_closed_conditional = self

        # Calculate the length of the body of the conditional
        # Loop in reverse through all pending_headers to calculate the lenght of all headers
        index = len(self._conn._pending_headers) - 1
        body_length = 0
        while self._conn._pending_headers[index] is not self.header:
            body_length += self._conn._pending_headers[index].HDR_LENGTH
            index -= 1
        
        # Set the correct length
        self.header.length = body_length
            
        # Set the scope to the parent scope
        self._conn.current_scope = self.parent



#
# Copyright (c) 2017, Stephanie Wehner and Axel Dahlberg
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
# 1. Redistributions of source code must retain the above copyright
#    notice, this list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright
#    notice, this list of conditions and the following disclaimer in the
#    documentation and/or other materials provided with the distribution.
# 3. All advertising materials mentioning features or use of this software
#    must display the following acknowledgement:
#    This product includes software developed by Stephanie Wehner, QuTech.
# 4. Neither the name of the QuTech organization nor the
#    names of its contributors may be used to endorse or promote products
#    derived from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDER ''AS IS'' AND ANY
# EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER BE LIABLE FOR ANY
# DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import os

from cqc.cqcHeader import CQCCmdHeader, CQC_CMD_MEASURE, CQC_CMD_MEASURE_INPLACE
from .qubit import qubit
from .cqc_handler import CQCHandler


class CQCToFile(CQCHandler):
    """Handler to be used when writing the CQC commands to a file."""

    def __init__(self, name=None, file='CQC_File', pend_messages=False,
                 overwrite=False, binary=True):

        if name is None:
            name = file

        # Call init of CQCHandler
        super().__init__(file, pend_messages=pend_messages)

        self.next_qubitID = 0

        self.binary = binary

        self.file = file

        # Check if file exists
        if overwrite:
            # Remove file if we can overwrite
            try:
                os.remove(self.file)
            except FileNotFoundError:
                pass
        else:
            if not os.path.isfile(self.file):
                pass
            else:
                # Append number to filename if can't overwrite
                num = 0
                while True:
                    if os.path.isfile(self.file + str(num)):
                        num += 1
                    else:
                        self.file = self.file + str(num)
                        break 

        # Don't want notify when writing to file
        self.notify = False

    def commit(self, msg):
        """Write a message to file.

        Message is written as string or as bytes depending on 
        self.binary
        """

        if self.binary is True:
            with open(self.file, 'ab') as f:
                f.write(msg)
        else:
            with open(self.file, 'a') as f:      
                f.write(str(msg) + '\n')

    def _handle_create_qubits(self, num_qubits):
        qubits = []
        for _ in range(num_qubits):
            q = qubit(self, createNew=False)
            q._qID = self.new_qubitID()
            q._set_active(True)
            qubits.append(q)

        return qubits

    def new_qubitID(self, print_cqc=False):
        """Provice new qubit ID.
        
        For CQCToFile we simply increase the qubit ID by one for each
        new qubit.
        """

        qubitID = self.next_qubitID

        self.next_qubitID += 1

        return qubitID

    def get_remote_from_directory_or_address(self, name):
        # Only return fixed address and port for now
        return 0, 0

    def _handle_epr_response(self, notify):
        # Initialize the qubit
        q = qubit(self, createNew=False)

        entInfoHdr = None  # TODO: create function that returns some fake entanglement info
        q_id = self.new_qubitID()

        q._set_entanglement_info(entInfoHdr)
        q._qID = q_id

        # Activate and return qubit
        q._set_active(True)

        return q

    def return_meas_outcome(self):
        """Return measurement outcome."""

        return 0

    def readMessage(self):
        """For now returns nothing"""
        return None

    def _handle_factory_response(self, num_iter, response_amount, should_notify=False):
        """Handles the responses from a factory command and returns a list of results"""
        res = []
        # Loop over the pending_headers to determine the total length and set should_notify
        for header in self._pending_headers:

            # Check if the current header is a Command header. It can also be a sub header
            if isinstance(header, CQCCmdHeader):
                
                if self.shouldReturn(header.instr):
                    # Build artificial responses
                    if header.instr in (CQC_CMD_MEASURE, CQC_CMD_MEASURE_INPLACE):
                        res.append(self.return_meas_outcome())
                    # TODO entanglement information etc
                    else:
                        q = None
                        if num_iter != 1:
                            q._set_active(False)
                            q = qubit(self, createNew=False)
                        if q is None:
                            q = qubit(self, createNew=False)
                        q._qID = self.new_qubitID()
                        q._set_entanglement_info(None)
                        q._set_active(True)
                        res.append(q)

        return res



# Copyright (c) 2017, Stephanie Wehner and Axel Dahlberg
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
# 1. Redistributions of source code must retain the above copyright
#    notice, this list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright
#    notice, this list of conditions and the following disclaimer in the
#    documentation and/or other materials provided with the distribution.
# 3. All advertising materials mentioning features or use of this software
#    must display the following acknowledgement:
#    This product includes software developed by Stephanie Wehner, QuTech.
# 4. Neither the name of the QuTech organization nor the
#    names of its contributors may be used to endorse or promote products
#    derived from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDER ''AS IS'' AND ANY
# EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER BE LIABLE FOR ANY
# DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import logging

from twisted.internet.defer import inlineCallbacks
from twisted.internet.protocol import Protocol, connectionDone

from cqc.cqcHeader import CQC_VERSION, CQCHeader

###############################################################################
#
# CQC Protocol
#
# Execute the CQC Protocol giving access to the SimulaQron backend via the
# universal interface.
#


class CQCProtocol(Protocol):
    # Dictionary storing the next unique qubit id for each used app_id
    _next_q_id = {}

    # Dictionary storing the next unique entanglement id for each used
    # (host_app_id,remote_node,remote_app_id)
    _next_ent_id = {}

    def __init__(self, factory):

        # CQC Factory, including our connection to the SimulaQron backend
        self.factory = factory

        # Default application ID, typically one connection per application but
        # we will deliberately NOT check for that since this is the task of
        # higher layers or an OS
        self.app_id = 0

        # Define the backend to use.
        self.messageHandler = factory.backend

        # Flag to determine whether we already received _all_ of the CQC header
        self.gotCQCHeader = False

        # Header for which we are currently processing a packet
        self.currHeader = None

        # Buffer received data (which may arrive in chunks)
        self.buf = None

        # Convenience
        self.name = self.factory.name

        logging.debug("CQC %s: Initialized Protocol", self.name)

    def connectionMade(self):
        pass

    def connectionLost(self, reason=connectionDone):
        pass

    def dataReceived(self, data):
        """
        Receive data. We will always wait to receive enough data for the
        header, and then the entire packet first before commencing processing.
        """
        # Read whatever we received into a buffer
        if self.buf:
            self.buf = self.buf + data
        else:
            self.buf = data

            # If we don't have the CQC header yet, try and read it in full.
        if not self.gotCQCHeader:
            if len(self.buf) < CQCHeader.HDR_LENGTH:
                # Not enough data for CQC header, return and wait for the rest
                return

                # Got enough data for the CQC Header so read it in
            self.gotCQCHeader = True
            raw_header = self.buf[0:CQCHeader.HDR_LENGTH]
            self.currHeader = CQCHeader(raw_header)

            # Remove the header from the buffer
            self.buf = self.buf[CQCHeader.HDR_LENGTH: len(self.buf)]

            logging.debug("CQC %s: Read CQC Header: %s", self.name, self.currHeader.printable())

            # Check whether we already received all the data
        if len(self.buf) < self.currHeader.length:
            # Still waiting for data
            logging.debug(
                "CQC %s: Incomplete data. Waiting. Current length %s, " "required length %s",
                self.name,
                len(self.buf),
                self.currHeader.length,
            )
            return

            # We got the header and all the data for this packet. Start processing.
            # Update our app ID
        self.app_id = self.currHeader.app_id
        # Invoke the relevant message handler, processing the possibly
        # remaining data
        try:
            self._parseData(self.currHeader, self.buf[0: self.currHeader.length])
        except Exception as e:
            print(e)
            import traceback

            traceback.print_exc()

            # if self.currHeader.tp in self.messageHandlers:
            # 	self.messageHandlers[self.currHeader.tp](self.currHeader, )
            # else:
            # 	self._send_back_cqc(self.currHeader, CQC_ERR_UNSUPP)

            # Reset and await the next packet
        self.gotCQCHeader = False

        # Check if we received data already for the next packet, if so save it
        if self.currHeader.length < len(self.buf):
            self.buf = self.buf[self.currHeader.length: len(self.buf)]
            self.dataReceived(b"")
        else:
            self.buf = None

    @inlineCallbacks
    def _parseData(self, header, data):
        try:
            yield self.messageHandler.handle_cqc_message(header, data)
            messages = self.messageHandler.retrieve_return_messages(header.app_id)
        except Exception as e:
            raise e

        if messages:
            # self.factory._lock.acquire()
            for msg in messages:
                self.transport.write(msg)
                # self.factory._lock.release()

    def _send_back_cqc(self, header, msgType, length=0):
        """
        Return a simple CQC header with the specified type.

        header	 CQC header of the packet we respond to
        msgType  Message type to return
        length	 Length of additional message
        """
        hdr = CQCHeader()
        hdr.setVals(CQC_VERSION, msgType, header.app_id, length)

        msg = hdr.pack()
        self.transport.write(msg)



#
# Copyright (c) 2017, Stephanie Wehner and Axel Dahlberg
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
# 1. Redistributions of source code must retain the above copyright
#    notice, this list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright
#    notice, this list of conditions and the following disclaimer in the
#    documentation and/or other materials provided with the distribution.
# 3. All advertising materials mentioning features or use of this software
#    must display the following acknowledgement:
#    This product includes software developed by Stephanie Wehner, QuTech.
# 4. Neither the name of the QuTech organization nor the
#    names of its contributors may be used to endorse or promote products
#    derived from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDER ''AS IS'' AND ANY
# EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER BE LIABLE FOR ANY
# DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import time
import socket
import logging

from cqc.cqcHeader import (
    CQC_TP_NEW_OK,
    CQC_TP_RECV,
    CQC_TP_EPR_OK,
    CQC_TP_MEASOUT,
    CQC_TP_EXPIRE,
    CQC_TP_INF_TIME,
    Header,
    CQCHeader,
    CQCMeasOutHeader,
    CQCTimeinfoHeader,
    CQCXtraQubitHeader,
)
from cqc.entInfoHeader import EntInfoHeader
from cqc.hostConfig import cqc_node_id_from_addrinfo
from .cqc_handler import CQCHandler
from .util import CQCUnsuppError
from .qubit import qubit

try:
    import simulaqron
    from simulaqron.general.hostConfig import socketsConfig
    from simulaqron.settings import simulaqron_settings
    _simulaqron_version = simulaqron.__version__
    _simulaqron_major = int(_simulaqron_version.split('.')[0])
except ImportError:
    _simulaqron_major = -1


class CQCConnection(CQCHandler):
    """Handler to be used when sending commands over a socket."""
    def __init__(self, name, socket_address=None, appID=None, pend_messages=False,
                 retry_connection=True, conn_retry_time=0.1, log_level=None, backend=None,
                 use_classical_communication=True, network_name=None):
        """
        Initialize a connection to the cqc server.

        Since version 3.0.0: If socket_address is None or use_classical_communication is True, the CQC connection
        needs some way of finding the correct socket addresses. If backend is None or "simulaqron" the connection
        will try to make use of the network config file setup in simulaqron. If simulaqron is not installed

        - **Arguments**
            :param name:        Name of the host.
            :param socket_address: tuple (str, int) of ip and port number.
            :param appID:        Application ID. If set to None, defaults to a nonused ID.
            :param pend_messages: True if you want to wait with sending messages to the back end.
                    Use flush() to send all pending messages in one go as a sequence to the server
            :param retry_connection: bool
                Whether to retry a failed connection or not
            :param conn_retry_time: float
                How many seconds to wait between each connection retry
            :param log_level: int or None
                The log-level, for example logging.DEBUG (default: logging.WARNING)
            :param backend: None or str
                If socket_address is None or use_classical_communication is True, If None or "simulaqron" is used
                the cqc library tries to use the network config file setup in simulaqron if network_config_file is None.
                If network_config_file is None and simulaqron is not installed a ValueError is raised.
            :param use_classical_communication: bool
                Whether to use the built-in classical communication or not.
            :param network_name: None or str
                Used if simulaqron is used to load socket addresses for the backend
        """

        super().__init__(
            name=name,
            app_id=appID,
            pend_messages=pend_messages,
        )

        self._setup_logging(log_level)

        # Connection retry time
        self._conn_retry_time = conn_retry_time

        # Buffer received data
        self.buf = None

        # ClassicalServer
        self._classicalServer = None

        # Classical connections in the application network
        self._classicalConn = {}

        # Get network configuraton and addresses
        addr, cqc_net, app_net = self._setup_network_data(
            socket_address=socket_address,
            use_classical_communication=use_classical_communication,
            backend=backend,
            network_name=network_name,
        )
        self._cqcNet = cqc_net
        self._appNet = app_net

        # Open a socket to the backend
        self._s = None
        cqc_socket = self._setup_socket(addr=addr, retry_connection=retry_connection)
        self._s = cqc_socket

    @staticmethod
    def _setup_logging(level):
        """
        Sets up the logging to the specified level (default logging.WARNING)
        :param level: int or None
            For example logging.DEBUG
        :return: None
        """
        if level is None:
            logging.basicConfig(format="%(asctime)s:%(levelname)s:%(message)s", level=logging.WARNING)
        else:
            logging.basicConfig(format="%(asctime)s:%(levelname)s:%(message)s", level=level)

    def _setup_network_data(self, socket_address, use_classical_communication, backend,
                            network_name):
        addr = None
        cqc_net = None
        app_net = None
        if socket_address is None or use_classical_communication:
            cqc_net, app_net = self._get_net_configs(
                use_classical_communication=use_classical_communication,
                backend=backend,
                network_name=network_name,
            )

            # Host data
            if self.name in cqc_net.hostDict:
                myHost = cqc_net.hostDict[self.name]
            else:
                raise ValueError("Host name '{}' is not in the cqc network".format(self.name))

                # Get IP and port number
            addr = myHost.addr

        if socket_address is not None:
            try:
                hostname, port = socket_address
                if not isinstance(hostname, str):
                    raise TypeError()
                if not isinstance(port, int):
                    raise TypeError()
                addrs = socket.getaddrinfo(hostname, port, proto=socket.IPPROTO_TCP, family=socket.AF_INET)
                addr = addrs[0]

            except Exception:
                raise TypeError("When specifying the socket address, this should be a tuple (str,int).")
        return addr, cqc_net, app_net

    def _get_net_configs(self, use_classical_communication, backend, network_name):
        cqc_net = None
        app_net = None
        if backend is None or backend == "simulaqron":
            if _simulaqron_major < 3:
                raise ValueError("If (socket_address is None or use_classical_communication is True)"
                                 "and (backend is None or 'simulaqron'\n"
                                 "you need simulaqron>=3.0.0 installed.")
            else:
                network_config_file = simulaqron_settings.network_config_file
                cqc_net = socketsConfig(network_config_file, network_name=network_name, config_type="cqc")
                if use_classical_communication:
                    app_net = socketsConfig(network_config_file, network_name=network_name, config_type="app")
        else:
            raise ValueError("Unknown backend")

        return cqc_net, app_net

    def _setup_socket(self, addr, retry_connection):
        cqc_socket = None
        while True:
            try:
                logging.debug("App {} : Trying to connect to CQC server".format(self.name))

                cqc_socket = socket.socket(addr[0], addr[1], addr[2])
                cqc_socket.connect(addr[4])
                break
            except ConnectionRefusedError as err:
                logging.debug("App {} : Could not connect to  CQC server, trying again...".format(self.name))
                time.sleep(self._conn_retry_time)
                cqc_socket.close()
                if not retry_connection:
                    self.close()
                    raise err
            except Exception as err:
                logging.exception("App {} : Critical error when connection to CQC server: {}".format(self.name, err))
                cqc_socket.close()
                raise err
        return cqc_socket

    def commit(self, msg):
        """Send message through the socket."""
        self._s.send(msg)

    def close(self, release_qubits=True, notify=True):
        """Handle closing actions.

        Flushes remaining headers, releases all qubits, closes the 
        connections, and removes the app ID from the used app IDs.
        """
        super().close()

        if self._s is not None:
            self._s.close()

        self.closeClassicalServer()

        for name in list(self._classicalConn):
            self.closeClassicalChannel(name)

    def new_qubitID(self, print_cqc=False):
        """Provide new qubit ID.
        
        For CQCConnection the qubit ID is given by the server. A message
        has to be read and the qubit ID extracted from it.
        """

        msg = self.readMessage()
        otherHdr = msg[1]

        if print_cqc:
            self.print_CQC_msg(msg)

        return otherHdr.qubit_id

    def startClassicalServer(self):
        """Sets up a server for the application communication, 
        if not already set up.
        """
        if self._appNet is None:
            raise ValueError(
                "Since use_classical_communication was set to False upon init, the built-in classical communication"
                "cannot be used."
            )

        if not self._classicalServer:
            logging.debug("App {}: Starting classical server".format(self.name))
            # Get host data
            myHost = self._appNet.hostDict[self.name]
            hostaddr = myHost.addr

            # Setup server
            s = socket.socket(hostaddr[0], hostaddr[1], hostaddr[2])
            s.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
            s.bind(hostaddr[4])
            s.listen(1)
            (conn, addr) = s.accept()
            logging.debug("App {}: Classical server started".format(self.name))
            self._classicalServer = conn

    def closeClassicalServer(self):
        """Closes classical server."""

        if self._classicalServer:
            logging.debug("App {}: Closing classical server".format(self.name))
            self._classicalServer.close()
            logging.debug("App {}: Classical server closed".format(self.name))
            self._classicalServer = None

    def recvClassical(self, timout=1, msg_size=1024, close_after=True):
        """Receive classical message."""

        if not self._classicalServer:
            self.startClassicalServer()
        for _ in range(10 * timout):
            logging.debug("App {}: Trying to receive classical message".format(self.name))
            msg = self._classicalServer.recv(msg_size)
            if len(msg) > 0:
                logging.debug("App {}: Received classical message".format(self.name))
                if close_after:
                    self.closeClassicalServer()
                return msg
            time.sleep(0.1)
        raise RuntimeError("Timeout: No message received")

    def openClassicalChannel(self, name):
        """
        Opens a classical connection to another host in the application network.

        - **Arguments**

            :name:        The name of the host in the application network.
            :timout:    The time to try to connect to the server. When timout is reached an RuntimeError is raised.
        """
        if self._appNet is None:
            raise ValueError(
                "Since use_classical_communication was set to False upon init, the built-in classical communication"
                "cannot be used."
            )
        if name not in self._classicalConn:
            logging.debug("App {}: Opening classical channel to {}".format(self.name, name))
            if name in self._appNet.hostDict:
                remoteHost = self._appNet.hostDict[name]
            else:
                raise ValueError("Host name '{}' is not in the cqc network".format(name))

            addr = remoteHost.addr
            while True:
                try:
                    s = socket.socket(addr[0], addr[1], addr[2])
                    s.connect(addr[4])
                    logging.debug("App {}: Classical channel to {} opened".format(self.name, name))
                    break
                except ConnectionRefusedError:
                    logging.debug(
                        "App {}: Could not open classical channel to {}, trying again..".format(self.name, name)
                    )
                    time.sleep(self._conn_retry_time)
                except Exception as e:
                    logging.warning(
                        "App {} : Critical error when connection to app node {}: {}".format(self.name, name, e)
                    )
                    break
            self._classicalConn[name] = s

    def closeClassicalChannel(self, name):
        """
        Closes a classical connection to another host in the application network.

        - **Arguments**

            :name:        The name of the host in the application network.
        """
        if name in self._classicalConn:
            logging.debug("App {}: Closing classical channel to {}".format(self.name, name))
            s = self._classicalConn.pop(name)
            s.close()
            logging.debug("App {}: Classical channel to {} closed".format(self.name, name))

    def sendClassical(self, name, msg, close_after=True):
        """
        Sends a classical message to another host in the application network.

        - **Arguments**

            :name:        The name of the host in the application network.
            :msg:        The message to send. Should be either a int in range(0,256) or a list of such ints.
            :timout:    The time to try to connect to the server. When timout is reached an RuntimeError is raised.
        """
        if name not in self._classicalConn:
            self.openClassicalChannel(name)
        try:
            to_send = bytes([int(msg)])
        except (TypeError, ValueError):
            to_send = bytes(msg)
        logging.debug("App {}: Sending classical message {} to {}".format(self.name, to_send, name))
        self._classicalConn[name].send(to_send)
        logging.debug("App {}: Classical message {} to {} sent".format(self.name, to_send, name))
        if close_after:
            self.closeClassicalChannel(name)

    def _handle_create_qubits(self, num_qubits, notify):
        qubits = []
        for _ in range(num_qubits):
            msg = self.readMessage()
            self.check_error(msg[0])
            if msg[0].tp != CQC_TP_NEW_OK:
                raise CQCUnsuppError("Unexpected message of type {} send back from backend".format(msg[0].tp))
            qubits.append(self.parse_CQC_msg(msg))
            self.print_CQC_msg(msg)

        if notify:
            message = self.readMessage()
            self._assert_done_message(message)
            self.print_CQC_msg(message)

        return qubits

    def readMessage(self, maxsize=192):  # WHAT IS GOOD SIZE?
        """Receive the whole message from cqc server.

        Returns (CQCHeader,None,None), (CQCHeader,CQCNotifyHeader,None) 
        or (CQCHeader,CQCNotifyHeader,EntInfoHeader) depending on the 
        type of message.

        Maxsize is the max size of message.
        """

        # Initialize checks
        gotCQCHeader = False
        if self.buf:
            checkedBuf = False
        else:
            checkedBuf = True

        while True:
            # If buf does not contain enough data, read in more
            if checkedBuf:
                # Receive data
                data = self._s.recv(maxsize)

                # Read whatever we received into a buffer
                if self.buf:
                    self.buf += data
                else:
                    self.buf = data

                    # If we don't have the CQC header yet, try and read it in full.
            if not gotCQCHeader:
                if len(self.buf) < CQCHeader.HDR_LENGTH:
                    # Not enough data for CQC header, return and wait for the rest
                    checkedBuf = True
                    continue

                    # Got enough data for the CQC Header so read it in
                gotCQCHeader = True
                rawHeader = self.buf[0:CQCHeader.HDR_LENGTH]
                currHeader = CQCHeader(rawHeader)

                # Remove the header from the buffer
                self.buf = self.buf[CQCHeader.HDR_LENGTH : len(self.buf)]

                # Check for error
                self.check_error(currHeader)

                # Check whether we already received all the data
            if len(self.buf) < currHeader.length:
                # Still waiting for data
                checkedBuf = True
                continue
            else:
                break
                # We got all the data, read other headers if there is any
        if currHeader.length == 0:
            return currHeader, None, None
        else:
            if currHeader.tp == CQC_TP_INF_TIME:
                timeinfo_header = self._extract_header(CQCTimeinfoHeader)
                return currHeader, timeinfo_header, None
            elif currHeader.tp == CQC_TP_MEASOUT:
                measout_header = self._extract_header(CQCMeasOutHeader)
                return currHeader, measout_header, None
            elif currHeader.tp in [CQC_TP_RECV, CQC_TP_NEW_OK, CQC_TP_EXPIRE]:
                xtra_qubit_header = self._extract_header(CQCXtraQubitHeader)
                return currHeader, xtra_qubit_header, None
            elif currHeader.tp == CQC_TP_EPR_OK:
                xtra_qubit_header = self._extract_header(CQCXtraQubitHeader)
                ent_info_hdr = self._extract_header(EntInfoHeader)
                return currHeader, xtra_qubit_header, ent_info_hdr

    def _extract_header(self, header_class):
        """
        Extracts the given header class from the first part of the current buffer.
        :param header_class: Subclassed from `cqc.backend.cqcHeader.Header`
        :return: An instance of the class
        """
        if not issubclass(header_class, Header):
            raise ValueError("header_class {} is not a subclass of Header".format(header_class))

        try:
            rawHeader = self.buf[:header_class.HDR_LENGTH]
        except IndexError:
            raise ValueError("Got a header message of unexpected size")
        self.buf = self.buf[header_class.HDR_LENGTH: len(self.buf)]
        header = header_class(rawHeader)

        return header

    def sendQubit(self, q, name, remote_appID=0, notify=True, block=True, remote_socket=None):
        """Sends qubit to another node in the cqc network. 
        
        If this node is not in the network an error is raised.

        - **Arguments**

            :q:         The qubit to send.
            :Name:         Name of the node as specified in the cqc network config file.
            :remote_appID:     The app ID of the application running on the receiving node.
            :nofify:     Do we wish to be notified when done.
            :block:         Do we want the qubit to be blocked
            :remote_socket: tuple (str, int) of ip and port number. Needed if no cqcFile was specified
        """
        q = super().sendQubit(
            q=q,
            name=name,
            remote_appID=remote_appID,
            notify=notify,
            block=block,
            remote_socket=remote_socket,
        )

    def createEPR(self, name, remote_appID=0, notify=True, block=True, remote_socket=None):
        """Creates epr with other host in cqc network.

        - **Arguments**

            :name:         Name of the node as specified in the cqc network config file.
            :remote_appID:     The app ID of the application running on the receiving node.
            :nofify:     Do we wish to be notified when done.
            :block:         Do we want the qubit to be blocked
            :remote_socket: tuple (str, int) of ip and port number. Needed if no cqcFile was specified
        """
        return super().createEPR(
            name,
            remote_appID=remote_appID,
            notify=notify,
            block=block,
            remote_socket=remote_socket,
        )

    def _handle_epr_response(self, notify):
        # Get RECV message
        message = self.readMessage()
        otherHdr = message[1]
        entInfoHdr = message[2]
        q_id = otherHdr.qubit_id

        self.print_CQC_msg(message)

        if notify:
            message = self.readMessage()
            self.print_CQC_msg(message)

        # initialize the qubit
        q = qubit(self, createNew=False)

        q._set_entanglement_info(entInfoHdr)
        q._qID = q_id
        # Activate and return qubit
        q._set_active(True)
        return q

    def _handle_factory_response(self, num_iter, response_amount, should_notify=False):
        """Handles the responses from a factory command and returns a list of results"""
        res = []
        for _ in range(num_iter):
            for _ in range(response_amount):
                message = self.readMessage()
                self.check_error(message[0])
                # TODO handle new qubit!
                res.append(self.parse_CQC_msg(message))
                self.print_CQC_msg(message)

        if should_notify:
            message = self.readMessage()
            self.check_error(message[0])

        return res

    def return_meas_outcome(self):
        """Return measurement outcome."""

        msg = self.readMessage()

        try:
            otherHdr = msg[1]
            return otherHdr.outcome
        except AttributeError:
            raise RuntimeError("Didn't receive a measurement outcome")

        message = self.readMessage()
        self._assert_done_message(message)
        self.print_CQC_msg(message)

    def get_remote_from_directory_or_address(self, name, remote_socket=None):
        cqcNet = self._cqcNet
        if remote_socket is None:
            try:
                # Get receiving host
                hostDict = cqcNet.hostDict
            except AttributeError:
                raise ValueError(
                    "If a CQCConnections is initialized without specifying a cqcFile you need to also provide a"
                    "socket address for the remote node here."
                )
            if name in hostDict:
                recvHost = hostDict[name]
                remote_ip = recvHost.ip
                remote_port = recvHost.port
            else:
                raise ValueError("Host name '{}' is not in the cqc network".format(name))
        else:
            try:
                remote_host, remote_port = remote_socket
                if not isinstance(remote_host, str):
                    raise TypeError()
                if not isinstance(remote_port, int):
                    raise TypeError()
            except Exception:
                raise TypeError("When specifying the remote socket address, this should be a tuple (str,int).")

                # Pack the IP
            addrs = socket.getaddrinfo(remote_host, remote_port, proto=socket.IPPROTO_TCP, family=socket.AF_INET)
            addr = addrs[0]
            remote_ip = cqc_node_id_from_addrinfo(addr)
            remote_port = addr[4][1]
        return remote_ip, remote_port



#
# Copyright (c) 2017, Stephanie Wehner and Axel Dahlberg
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
# 1. Redistributions of source code must retain the above copyright
#    notice, this list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright
#    notice, this list of conditions and the following disclaimer in the
#    documentation and/or other materials provided with the distribution.
# 3. All advertising materials mentioning features or use of this software
#    must display the following acknowledgement:
#    This product includes software developed by Stephanie Wehner, QuTech.
# 4. Neither the name of the QuTech organization nor the
#    names of its contributors may be used to endorse or promote products
#    derived from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDER ''AS IS'' AND ANY
# EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER BE LIABLE FOR ANY
# DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

from cqc.pythonLib import qubit


def parity_meas(qubits, bases, node, negative=False):
    """
    Performs a parity measurement on the provided qubits in the Pauli bases specified by 'bases'.
    'bases' should be a string with letters in "IXYZ" and have the same length as the number of qubits provided.
    If 'negative' is true the measurement outcome is flipped.
    If more than one letter of 'bases' is not identity, then an ancilla qubit will be used, which is created using the
    provided 'node'.

    :param qubits: List of qubits to be measured.
    :type qubits: list of :obj: `cqc.pythonLib.qubit`
    :param bases: String specifying the Pauli-bases of the measurement. Example bases="IXY" for three qubits.
    :type bases: str
    :param node: The node storing the qubits. Used for creating an ancilla qubit.
    :type node: :obj: `cqc.pythonLib.CQCConnection`
    :param negative: If the measurement outcome should be flipped or not.
    :type negative: bool
    :return: The measurement outcome 0 or 1, where 0 correspond to the +1 eigenvalue of the measurement operator.
    """

    if not (len(qubits) == len(bases)):
        raise ValueError("Number of bases needs to be the number of qubits.")
    if not all([(B in "IXYZ") for B in bases]):
        raise ValueError("All elements of bases need to be in 'IXYZ'.")

    num_qubits = len(qubits)

    flip_basis = ["I"] * num_qubits
    non_identity_bases = []

    # Check if we need to flip the bases of the qubits
    for i in range(len(bases)):
        B = bases[i]
        if B == "X":
            flip_basis[i] = "H"
            non_identity_bases.append(i)
        elif B == "Y":
            flip_basis[i] = "K"
            non_identity_bases.append(i)
        elif B == "Z":
            non_identity_bases.append(i)
        else:
            pass

    if len(non_identity_bases) == 0:
        # Trivial measurement
        m = 0

    elif len(non_identity_bases) == 1:
        # Single_qubit measurement
        q_index = non_identity_bases[0]
        q = qubits[q_index]

        # Flip to correct basis
        if flip_basis[q_index] == "H":
            q.H()
        if flip_basis[q_index] == "K":
            q.K()

        m = q.measure(inplace=True)

        # Flip the qubit back
        if flip_basis[q_index] == "H":
            q.H()
        if flip_basis[q_index] == "K":
            q.K()

    else:
        # Parity measurement, ancilla needed

        # Initialize ancilla qubit
        anc = qubit(node)

        # Flip to correct basis
        for i in range(len(bases)):
            if flip_basis[i] == "H":
                qubits[i].H()
            if flip_basis[i] == "K":
                qubits[i].K()

                # Transfer parity information to ancilla qubit
        for i in non_identity_bases:
            qubits[i].cnot(anc)

            # Measure ancilla qubit
        m = anc.measure()

        # Flip to correct basis
        for i in range(len(bases)):
            if flip_basis[i] == "H":
                qubits[i].H()
            if flip_basis[i] == "K":
                qubits[i].K()
    if negative:
        return (m + 1) % 2
    else:
        return m
